from __future__ import division

import pygame

from random import random,randrange

from util import load_image

class Animation:
    def __init__(self,data):
        self.data = data
        self.frame = 0
        self.time = 0
        self.delay = 0
        self.playing = False
        self.loop = False
        self.image = None
    def play(self,loop = True,delay = 0):
        self.delay = delay
        self.playing = True
        self.frame = 0
        self.image,self.time = self.data[self.frame]
    def draw(self,screen,pos,offset):
        x,y = pos
        ox,oy = offset
        if self.delay:
            self.delay = max(self.delay-1,0)
        else:
            if self.time > 0:
                self.time -= 1
            elif self.time == 0:
                size = len(self.data)
                if self.loop:
                    self.frame += 1
                    self.image,self.time = self.data[self.frame%size]
                else:
                    if self.frame < size-1:
                        self.frame += 1
                        self.image,self.time = self.data[self.frame]
                    else:
                        self.playing = False
                        self.anim = None
                        self.frame = self.time = 0
                        return True
            i = load_image(self.image)
            w,h = i.get_size()
            screen.blit(i,(ox+x-w/2,oy+y-h/2))

class Fountain:
    def __init__(self,colour,pos):
        self.image = load_image("blood-%s"% colour)
        self.pos = pos
        self.particles = []
        self.dead = False
        self.time = 0
    def draw(self,screen,offset):
        x,y = self.pos
        ox,oy = offset
        if self.time == 0:
            for i in range(28):
                px = randrange(-20,20)+x
                py = randrange(-20,20)+y
                self.particles.append((px,py,(randrange(-10,10)/10),randrange(120,160)/10))
        if self.time < 60:
            for i in range(3):
                px = randrange(-20,20)+x
                py = randrange(-20,20)+y
                self.particles.append((px,py,(randrange(-10,10)/10),randrange(120,160)/10))
        for i,p in enumerate(self.particles):
            x,y,dx,dy = p
            x += dx
            y -= dy
            dy -= 0.4
            screen.blit(self.image,(x+ox,y+oy))

            p = (x,y,dx,dy)
            if y+oy < 540:
                self.particles[i] = p
            else:
                self.particles.pop(i)
        self.time += 1
        if not self.particles:
            self.dead = True
        

class LineEffect:
    def __init__(self,anim,pos,dest,time,loop=False,delay=0):
        self.anim = Animation(anim)
        self.pos = pos
        x,y = pos
        dx,dy = dest
        t = time
        self.v = ((dx-x)/t,(dy-y)/t)
        self.time = time
        self.anim.play(loop,delay)
        self.dead = False
    def draw(self,screen,offset):
        x,y = self.pos
        dx,dy = self.v
        x += dx
        y += dy
        self.pos = (x,y)
        self.time -= 1
        if not self.dead:
            self.dead = self.anim.draw(screen,self.pos,offset)
        if self.time == 0:
            self.dead = True

