import pygame

from pygame.locals import *

import pickle

from effect import Fountain
from gui import *
from scene import *
from util import *

import map

TYPEA = {
    K_UP:(0,-1),
    K_DOWN:(0,1),
    K_LEFT:(-1,0),
    K_RIGHT:(1,0),

    K_w:(0,-1),
    K_s:(0,1),
    K_a:(-1,0),
    K_d:(1,0),
    
    K_KP8:(0,-1),
    K_KP2:(0,1),
    K_KP4:(-1,0),
    K_KP6:(1,0),
    K_KP7:(-1,-1),
    K_KP9:(1,-1),
    K_KP1:(-1,1),
    K_KP3:(1,1),
    }

TYPEB = {
    K_LEFT:(0,1),
    K_RIGHT:(0,-1),
    K_DOWN:(1,0),
    K_UP:(-1,0),

    K_a:(0,1),
    K_d:(0,-1),
    K_s:(1,0),
    K_w:(-1,0),

    K_KP8:(-1,0),
    K_KP2:(1,0),
    K_KP4:(0,1),
    K_KP6:(0,-1),
    K_KP7:(-1,1),
    K_KP9:(-1,-1),
    K_KP1:(1,1),
    K_KP3:(1,-1),
    }

colours = {
        "red":Red_Jelly,
        "blue":Blue_Jelly,
        "yellow":Yellow_Jelly,
        
    ##    "purple":Purple_Jelly,
    ##    "orange":Orange_Jelly,
        "green":Green_Jelly,
        }

class Core:
    def __init__(self,screen):
        self.screen = screen
        self.states = {
            "battle":BattleState(self),
            }
        self.clock = pygame.time.Clock()
        self.c_state = self.states["battle"]
    def run(self):
        self.c_state.run()

class BattleState:
    def __init__(self,core):
        self.core = core
        self.screen = core.screen
        self.type = TYPEA
        self.arrow = Arrow(self.screen)
        self.panel = StatView(self.screen)

        self.load_level("001")
    def load_level(self,level):
        self.reset()

        enemies = {
            "grey":Grey_Ooze,
            }
        
        lv = open("data/levels/level-%s.txt"% level)
        L = pickle.load(lv)
        self.floor = map.load_map("map-%s"%level,self)
        for c,m,pos in L["enemies"]:
            s = self.floor.data[pos]
            unit = enemies[c]((pos[0],pos[1],s.height()),m)
            s.add(unit)
            self.enemies.append(unit)
        self.camera.look_at(self.floor.cursor,25)
    def reset(self):
        self.camera = Camera(0,0)
        self.floor = None
        self.allies = []
        self.enemies = []
        self.effects = []
        self.phase = "free"
        self.selected = None
        self.highlights = {}
        self.undo = None
        self.side = "jelly"
        self.turn = 0
        self.jelly = 0
    def draw(self):
        self.screen.fill((0,0,0))
        self.floor.draw()
        self.arrow.draw(self.camera.offset())
        for fx in self.effects:
            if fx.dead:
                self.effects.remove(fx)
                continue
            fx.draw(self.screen,self.camera.offset())
        x,y = (self.floor.cursor.x,self.floor.cursor.y)
        u = None
        for t in self.floor.data[(x,y)].objects.__reversed__():
            if t.type == "unit":
                u = t
                break
        if u:
            self.panel.draw(u,(30,15))            
        pygame.display.flip()
    def end_turn(self):
        self.selected = self.arrow.target = None
        self.turn += 1
        self.side = "jelly"
        self.floor.cursor.show(self.floor)
        self.phase = "free"
        for a in self.allies:
            a.tired = False
        for e in self.enemies:
            e.tired = False
        self.camera.look_at(self.floor.cursor,10)
    def run(self):
        clock = self.core.clock
        m = self.floor
        c = m.cursor
        f = self.selected
        targets = []
        mixing = False
        reagent = None
        while True:
            clock.tick(60)
            self.draw()
            event = pygame.event.poll()
            keys = pygame.key.get_pressed()
            if event.type == QUIT:
                raise SystemExit()

# FREE STATE
            if self.phase == "free":
                if event.type == KEYDOWN:
                    if event.key in (K_x,K_RETURN,K_SPACE,K_KP5):
                        o = c.select(m)
                        if o.type == "unit":
                            self.arrow.target = f = o
                            if o.side == "jelly" and o.tired == False:
                                self.phase = "move"
                        elif o.type == "spawn":
                            s = self.floor.data[(c.x,c.y)]
                            unit = colours[o.colour]((c.x,c.y,s.height()))
                            s.add(unit)
                            self.allies.append(unit)
                            c.move((0,0),m)
                        else:
                            f = self.arrow.target = None
                    elif event.key == K_t:
                        c.hide(m)
                        self.phase = "ai"
                        self.side = "ooze"
                        continue
                for k in self.type:
                    if keys[k]:
                        c.move(self.type[k],m)
                        self.camera.look_at(c,8)

# MOVEMENT DECISION STATE
            elif self.phase == "move":
                # Find walkable area
                if not self.highlights:
                    self.highlights = f.walkable_tiles(m)
                    # Add indicator panels
                    for x,y in self.highlights:
                        px,py = (f.x,f.y)
                        s = m.data[(x,y)]
                        if x != px or y != py:
                            if s.top().type == "unit":
                                s.add(Panel("purple",(x,y,s.height())))
                            else:
                                s.add(Panel("blue",(x,y,s.height())))
                        else:
                            s.add(Panel("wait",(x,y,s.height())))
                # Check for input
                for k in self.type:
                    if keys[k]:
                        c.move(self.type[k],m)
                        self.camera.look_at(c,8)
                    if event.type == KEYDOWN:
                        if event.key in (K_x,K_RETURN,K_SPACE,K_KP5):
                            x,y = (c.x,c.y)
                            if x == f.x and y == f.y:
                                for x,y in self.highlights:
                                    m.data[(x,y)].delight()
                                c.move((0,0),m)
                                self.highlights = {}
                                self.undo = (f.x,f.y,f.z)
                                self.phase = "attack"
                            elif m.data[(x,y)].indicated():
                                if m.data[(x,y)].indicated("purple"):
                                    mixing = True
                                    reagent = m.data[(x,y)].top()
                                movement = self.highlights[(x,y)]
                                f.path = movement[1:]
                                for x,y in self.highlights:
                                    m.data[(x,y)].delight()
                                c.move((0,0),m)
                                self.highlights = {}
                                self.undo = (f.x,f.y,f.z)
                                self.phase = "moving"
                                c.hide(m)
                        elif event.key in (K_z,K_BACKSPACE):
                            for x,y in self.highlights:
                                m.data[(x,y)].delight()
                            c.move((0,0),m)
                            self.highlights = {}
                            self.arrow.target = f = None
                            self.phase = "free"

# ENEMY AI STAGE
            elif self.phase == "ai":
                enemy = None
                for e in self.enemies:
                    if e.tired == False:
                        enemy = e
                        break
                if enemy:
                    f = self.arrow.target = enemy
                    f.ai(m)
                    self.phase = "moving"
                else:
                    self.end_turn()

# MOVEMENT STAGE
            elif self.phase == "moving":
                if f.move(m):
                    self.camera.look_at(f,16)
                else:
                    if mixing:
                        self.allies.remove(f)
                        self.allies.remove(reagent)
                        self.allies.append(mix(f,reagent,m))
                        reagent = None
                        mixing = False
                        f = self.arrow.selected = None
                        self.phase = "free"
                    else:
                        self.phase = "attack"
                    c.show(m)

# ATTACK DECISION STATE
            elif self.phase == "attack":
                if self.side == "jelly":
                    # Find attackable squares
                    if not self.highlights:
                        self.highlights = f.attack_tiles(m,(f.x,f.y))
                        # Add indicator panels
                        for x,y in self.highlights:
                            px,py = (f.x,f.y)
                            s = m.data[(x,y)]
                            if x != px or y != py:
                                s.add(Panel("red",(x,y,s.height())))
                            else:
                                s.add(Panel("wait",(x,y,s.height())))
                                
                    # Check for input
                    for k in self.type:
                        if keys[k]:
                            c.move(self.type[k],m)
                            self.camera.look_at(c,8)
                            # Remove yellow targets
                            for x,y in targets:
                                m.data[(x,y)].delight(colour="yellow")
                            # Set yellow targets
                            targets = f.target_tiles(m,(c.x,c.y))
                            for x,y in targets:
                                s = m.data[(x,y)]
                                s.add(Panel("yellow",(x,y,s.height())))
                    if event.type == KEYDOWN:
                        if event.key in (K_x,K_RETURN,K_SPACE,K_KP5):
                            s = m.data[(c.x,c.y)]
                            if c.x == f.x and c.y == f.y:
                                for x,y in self.highlights:
                                    m.data[(x,y)].delight()
                                for x,y in targets:
                                    m.data[(x,y)].delight(colour="yellow")
                                targets = []
                                self.highlights = {}
                                self.phase = "free"
                                c.move((0,0),m)
                                f.tired = True
                                self.arrow.target = f = None
                            elif s.indicated("red"):
                                for x,y in self.highlights:
                                    m.data[(x,y)].delight()
                                for x,y in targets:
                                    m.data[(x,y)].delight(colour="yellow")
                                self.highlights = {}
                                self.phase = "combat"
                        if event.key in (K_z,K_BACKSPACE):
                            for x,y in self.highlights:
                                m.data[(x,y)].delight()
                            for x,y in targets:
                                m.data[(x,y)].delight(colour="yellow")
                            targets = []
                            self.highlights = {}
                            self.phase = "free"
                            f.reset(self.undo,m)
                            self.camera.look_at(f,10)
                            c.move((f.x-c.x,f.y-c.y),m)
                            self.arrow.target = f = None
                elif self.side == "ooze":
                    targets = f.target_tiles(m,f.target)
                    self.phase = "combat"
# COMBAT STATE
            elif self.phase == "combat":
                if f:
                    c.hide(m)
                    victims = find_targets(targets,m)
                    data = f.action(targets,victims,(c.x,c.y),m)
                    for fx in data:
                        self.effects.append(fx)
                    a = f.attack
                    f.tired = True
                    self.arrow.target = f = None
                if self.effects == []:
                    for v in victims[:]:
                        d = v.hurt(a)
                        if d:
                            x,y = v.pos
                            self.effects.append(Fountain(v.colour,(x,y-32)))
                            m.data[(v.x,v.y)].remove(v)
                            victims.remove(v)
                            if v.side == "jelly":
                                self.allies.remove(v)
                            elif v.side == "ooze":
                                self.enemies.remove(v)
                    victims = None
                    data = None
                    if self.side == "jelly":
                        self.phase = "free"
                        c.show(m)
                    else:
                        self.phase = "ai"

            self.camera.tick()
            m.cursor.tick()
