from __future__ import division

from math import hypot, sin

import pygame

from effect import *
from util import *

from random import choice, random



def miscable(j1,j2):
    mix_table = {
        "red":Red_Jelly,
        "blue":Blue_Jelly,
        "yellow":Yellow_Jelly,
        
    ##    "purple":Purple_Jelly,
    ##    "orange":Orange_Jelly,
        "green":Green_Jelly,

    ##    "pink":Pink_Jelly,
    ##    "grape":Grape_Jelly,
    ##    "lime":Lime_Jelly,
    ##    "lemon":Lemon_Jelly,
    ##    "ochre":Ochre_Jelly,
    ##    "olive":Olive_Jelly,
        

    ##    ("red","yellow"):Orange_Jelly,
    ##    ("blue","red"):Purple_Jelly,
        ("blue","yellow"):Green_Jelly,

        
    ##    ("purple","red"):Pink_Jelly,
    ##    ("blue","purple"):Grape_Jelly,
    ##    ("green","yellow"):Lime_Jelly,
    ##    ("orange","yellow"):Lemon_Jelly,
    ##    ("orange","red"):Ochre_Jelly,
    ##    ("blue","green"):Olive_Jelly,
        }
    
    jelly = None
    if j1.side == j2.side == "jelly":
        if j1.colour == j2.colour:
            jelly = mix_table.get(j1.colour)
        else:
            j = [j1.colour,j2.colour]
            j.sort()
            jelly = mix_table.get((j[0],j[1]))
    return jelly

def mix(j1,j2,floor):
    jelly = miscable(j1,j2)
    hp = j1.hp + j2.hp

    x = j1.x
    y = j1.y

    floor.data[(x,y)].remove(j1)
    floor.data[(x,y)].remove(j2)

    z = floor.data[(x,y)].height()

    unit = jelly((x,y,z))
    unit.hp = min(unit.max_hp,hp)
    floor.data[(x,y)].add(unit)
    return unit

class Camera:
    def __init__(self,x,y):
        self.pos = (x,y)
        self.motion = None
        self.time = 0
    def look_to(self,pos,time):
        """Looks to a point specified."""
        x,y = self.pos
        nx,ny = pos
        dx = (nx - x)/time
        dy = (ny - y)/time
        self.motion = (dx,dy)
        self.time = time
    def look_at(self,obj,time):
        """Looks at an object with a pos value."""
        self.look_to(obj.pos,time)
    def move(self,x,y):
        px,py = self.pos
        px += x
        py += y
        self.pos = (px,py)
    def offset(self):
        x,y = self.pos
        ox,oy = (-x+500,-y+260)
        return (ox,oy)
    def tick(self):
        if self.motion:
            x,y = self.pos
            mx,my = self.motion
            x += mx
            y += my
            self.time -= 1
            self.pos = (x,y)
            if self.time == 0:
                self.motion = None

class Stack:
    def __init__(self,pos):
        self.x,self.y = pos
        self.pos = (self.x*32-self.y*32,(self.x+self.y)*16)
        self.objects = []
    def draw(self,screen,offset):
        for o in self.objects:
            o.draw(screen,offset)
    def height(self):
        h = 0
        for o in self.objects:
            h += o.h
        return h
    def indicated(self,colour = None):
        for o in self.objects:
            if o.type == "indicator":
                return colour is None or colour == o.colour
        return False
    def top(self):
        for o in self.objects.__reversed__():
            if o.type in ("indicator","cursor"):
                continue
            else:
                return o
    def add(self,item,cur = False):
        if cur:
            if self.objects[-1].type not in ("floor","indicator","spawn"):
                item.z = self.height()-self.objects[-1].h
                self.objects.insert(-1,item)
            else:
                item.z = self.height()
                self.objects.append(item)                
        else:
            self.objects.append(item)
    def remove(self,item):
        self.objects.remove(item)
    def delight(self,colour = None):
        for o in self.objects:
            if colour:
                if o.type == "indicator" and o.colour == colour:
                    self.objects.remove(o)
            else:
                if o.type == "indicator":
                    self.objects.remove(o)
    def __getitem__(self,index):
        return self.objects[index]

class Cursor:
    def __init__(self,image,pos):
        self.image = load_image(image)
        self.type = "cursor"
        self.x,self.y = pos
        self.z = 0
        self.pos = (self.x*32-self.y*32,(self.x+self.y)*16-self.z)
        self.h = 0
        self.rect = self.image.get_rect()
        self.cd = 0
    def draw(self,screen,offset):
        ox,oy = offset
        px,py = self.pos
        x = px+ox
        y = py+oy
        self.rect.midbottom = (x,y)
        screen.blit(self.image,self.rect)
        self.pos = (self.x*32-self.y*32,(self.x+self.y)*16-self.z)
    def move(self,d,floor):
        if self.cd == 0:
            x,y = (self.x,self.y)
            dx,dy = d
            floor.data[(x,y)].remove(self)
            x = max(min(dx+x,floor.length-1),0)
            y = max(min(dy+y,floor.length-1),0)
            floor.data[(x,y)].add(self,cur=True)
            self.x = x
            self.y = y
            self.cd += 8
    def hide(self,floor):
        s = floor.data[(self.x,self.y)]
        if self in s.objects:
            s.remove(self)
    def show(self,floor):
        s = floor.data[(self.x,self.y)]
        if self not in s.objects:
            s.add(self,cur=True)
    def select(self,floor):
        s = floor.data[(self.x,self.y)].top()
        return s
    def tick(self):
        self.cd = max(self.cd-1,0)

class Arrow:
    def __init__(self,screen):
        self.screen = screen
        self.target = None
        self.images = {
            "jelly":load_image("arrow-ally"),
            "ooze":load_image("arrow-enemy"),
            None:load_image("arrow-object"),
            }
        self.cycle = 0
        self.rect = self.images["jelly"].get_rect()
    def draw(self,offset):
        self.cycle += 0.1
        ox,oy = offset
        if self.target:
            i = self.images[self.target.side]
            x,y = self.target.pos
            self.rect.midbottom = (ox+x,oy+y-self.target.h-10 + sin(self.cycle)*6)
            self.screen.blit(i,self.rect)

class Panel:
    def __init__(self,colour,pos):
        self.x,self.y,self.z = pos
        self.h = 6
        self.pos = (self.x*32-self.y*32,(self.x+self.y)*16-self.z)
        self.image = load_image("panel-%s"% colour)
        self.colour = colour
        self.type = "indicator"
        self.rect = self.image.get_rect()
        self.walkable = False
    def draw(self,screen,offset):
        ox,oy = offset
        px,py = self.pos
        x = px+ox
        y = py+oy
        self.rect.midbottom = (x,y)
        screen.blit(self.image,self.rect)

class Tile:
    def __init__(self,image,pos,height,walkable):
        self.x,self.y,self.z = pos
        self.h = height
        self.pos = (self.x*32-self.y*32,(self.x+self.y)*16-self.z)
        self.image = load_image(image)
        self.type = "floor"
        self.rect = self.image.get_rect()
        self.walkable = walkable
    def draw(self,screen,offset):
        ox,oy = offset
        px,py = self.pos
        x = px+ox
        y = py+oy
        self.rect.midbottom = (x,y)
        screen.blit(self.image,self.rect)

class Spawn:
    def __init__(self,colour,pos):
        self.x,self.y,self.z = pos
        self.h = 10
        self.pos = (self.x*32-self.y*32,(self.x+self.y)*16-self.z)
        self.image = load_image("spawnpoint")
        self.ball = load_image("blood-%s"%colour)
        self.colour = colour
        self.type = "spawn"
        self.rect = self.image.get_rect()
        self.walkable = True
        self.time = 0
    def draw(self,screen,offset):
        self.time += 1
        ox,oy = offset
        px,py = self.pos
        x = px+ox
        y = py+oy
        self.rect.midbottom = (x,y)
        screen.blit(self.image,self.rect)
        screen.blit(self.ball,(x-4,y-40+8*sin(self.time/3)))
    
class Unit:
    colour = "blue"
    side = "jelly"
    
    max_hp = 100
    attack = 20
    defense = 10

    moverange = 5
    attackrange = 1
    attackmin = 0
    targetsize = 0.5
    targetmin = 0
    mode = "circle"
    shape = "circle"
    jump = 24
    
    def __init__(self,pos,personality="lazy"):
        self.x,self.y,self.z = pos
        self.pos = (self.x*32-self.y*32,(self.x+self.y)*16-self.z)
        self.h = 24
        ipath = "%s-%s"%(self.side,self.colour)
        self.images = {
            "n":load_image("%s-n"% ipath),
            "s":load_image("%s-s"% ipath),
            "e":load_image("%s-e"% ipath),
            "w":load_image("%s-w"% ipath),
            }
        moods = {
            "lazy":self.lazy,
            }
        self.personality = moods[personality]
        self.facing = choice(self.images.keys())
        self.image = self.images[self.facing]
        self.sleepy = load_image("tired")
        self.type = "unit"
        self.rect = self.image.get_rect()
        self.walkable = False
        
        self.path = []
        self.target = None
        
        self.switchd = False
        self.v = None
        self.frames = 0

        self.hp = self.max_hp
        
        self.damage = 0
        self.hurtframes = 0
        
        self.jumping = False
        self.tired = False

    def draw(self,screen,offset):
        ox,oy = offset
        px,py = self.pos
        x = px+ox
        y = py+oy-12
        self.rect.midbottom = (x,y)
        if self.path and not self.jumping:
            s = sin(self.frames/2.55)
            i = pygame.transform.scale(self.image,(int(64+3*s),int(64-6*s)))
            screen.blit(i,self.rect)
        else:
            screen.blit(self.image,self.rect)
        if self.tired:
            u,v = (x+6,y-self.h)
            screen.blit(self.sleepy,(u,v))
        if self.hurtframes:
            self.hurtframes -= 1
            if self.damage < 0:
                c = "green"
            else:
                c = "red"
            ybuff = sin((3.14/30)*self.hurtframes)
            n = 0
            for d in str(abs(self.damage)):
                screen.blit(load_image("%s-%s"%(d,c)),(x-10+n*8,y-self.h-16-ybuff*5))
                n += 1
    def ai(self,floor):
        actions = []
        best_action = None
        brain = self.personality
        paths = self.walkable_tiles(floor)
        for p in paths:
            reach = self.attack_tiles(floor,p)
            for t in reach:
                actions.append(brain(paths[p],t,floor))
        best_action = actions[0]
        
        for a in actions:
            c = [best_action,a]
            c.sort()
            best_action = c[-1]

        self.path = best_action[1][1:]
        self.target = best_action[2]
                
    def action(self,targets,victims,cursor,floor):
        return []
    def hurt(self,attack):
        if attack >= 0:
            bd = (attack/self.defense)*10
        else:
            bd = attack
        d = bd*(random()+0.5)
        self.damage = int(d)
        self.hp = max(min(self.hp-self.damage,self.max_hp),0)
        self.hurtframes = 30

        if self.hp == 0:
            return True
    def reset(self,pos,floor):
        ux,uy,uz = pos
        floor.data[(self.x,self.y)].remove(self)
        self.x = ux
        self.y = uy
        self.z = uz
        floor.data[(self.x,self.y)].add(self)
        self.pos = (self.x*32-self.y*32,(self.x+self.y)*16-self.z)
    def column(self,ownstack,dstack):
        self.x = dstack[-1].x
        self.y = dstack[-1].y
        self.z = dstack.height()
        ownstack.remove(self)
        dstack.add(self)
    def move(self,floor):
        if self.path:
            ownstack = floor.data[(self.x,self.y)]
            dstack = floor.data[(self.path[0])]
            tx,ty = dstack[-1].pos
            ty -= (dstack[-1].h)
            x,y = self.pos
            if self.v:
                x += self.v[0]
                y += self.v[1]
                self.frames -= 1
                if self.jumping:
                    g = (1/8)-(1/64)*self.frames
                    y += g*(self.jump*2)
                self.pos = (x,y)
            else:
                self.v = ((tx-x)/16,(ty-y)/16)
                f = get_facing(self.v)
                if f:
                    self.facing = f
                    self.image = self.images[self.facing]
                self.frames = 16
                if dstack.height() != ownstack.height()-self.h:
                    self.jumping = True
                else:
                    self.jumping = False
            if self.frames == 16:
                dx,dy = self.path[0]
                if dx > self.x or dy > self.y:
                    self.column(ownstack,dstack)
                    self.switchd = True
                else:
                    self.switchd = False
            elif self.frames == 0:
                if not self.switchd:
                    self.column(ownstack,dstack)
                self.path.pop(0)
                self.v = None
                self.pos = (self.x*32-self.y*32,(self.x+self.y)*16-self.z)
            return True
        else:
            return False

    def attack_tiles(self,floor,centre):
        targets = []
        cx,cy = centre
        if self.mode == "circle":
            for y in range(int(-self.attackrange),int(self.attackrange+1)):
                for x in range(int(-self.attackrange),int(self.attackrange+1)):
                    if self.attackmin <= hypot(x,y) <= self.attackrange:
                        if floor.data.has_key((cx+x,cy+y)):
                            targets.append((cx+x,cy+y))
        return targets
    def target_tiles(self,floor,centre):
        targets = []
        cx,cy = centre
        if self.shape == "circle":
            for y in range(int(-self.targetsize),int(self.targetsize+1)):
                for x in range(int(-self.targetsize),int(self.targetsize+1)):
                    if self.targetmin <= hypot(x,y) <= self.targetsize:
                        if floor.data.has_key((cx+x,cy+y)):
                            targets.append((cx+x,cy+y))
        return targets
        
    def walkable_tiles(self,floor):
        stacks = floor.data
        working = [[(self.x,self.y)]]
        tiles = {}
        tiles[(self.x,self.y)] = []
        while working:
            poslist = working.pop(0)
            x,y = poslist[-1]
            adj = [poslist+[(x+i,y+j)]
                   for i,j in [(-1,0),(0,-1),(0,1),(1,0)]]
            for newlist in adj:
                ax,ay = newlist[-1]
                steps = len(newlist)
                if (ax,ay) in tiles or steps > self.moverange:
                    continue
                surf = stacks[(x,y)].top()
                dest = stacks.get((ax,ay))
                if dest:
                    if dest.top().type == "unit" and miscable(self,dest.top()):
                        working.append(newlist)
                        tiles[(ax,ay)] = newlist
                    elif dest.top().walkable:
                        if surf == self:
                            z = stacks[(x,y)].height()-self.h
                        else:
                            z = stacks[(x,y)].height()
                        tz = dest.top().z+dest.top().h
                        dz = abs(tz-z)
                        if dz <= self.jump:
                            working.append(newlist)
                            tiles[(ax,ay)] = newlist
        return tiles

    def net_damage(self,targets,floor):
        d = 0
        for t in targets:
            if t.side == self.side:
                d -= self.attack
            else:
                d += self.attack
        return d
    def lazy(self,path,target,floor):
        x,y = (self.x,self.y)
        tx,ty = target
        areas = self.target_tiles(floor,target)
        victims = find_targets(areas,floor)
        d = self.net_damage(victims,floor)
        dist = hypot(tx-x,ty-y)
        return ((d,-dist),path,target)
        

class Blue_Jelly(Unit):
    name = "Blue Jelly"
    colour = "blue"
    side = "jelly"
    
    max_hp = 50
    attack = 20
    defense = 8

    moverange = 5
    attackrange = 3
    targetsize = 1
    mode = "circle"
    shape = "circle"
    jump = 24
    def action(self,targets,victims,cursor,floor):
        cx,cy = cursor
        t = floor.data[(cx,cy)][-1]
        x,y = t.pos
        effects = [LineEffect([
                ("beam1",3),
                ("beam2",4),
                ("beam3",5),
                ("beam4",6),
                ("beam5",5),
                ("beam6",4),
                ],(x,y-500),(x,y-16),20),
                   
                    LineEffect([
                ("nova01",2),
                ("nova02",2),
                ("nova03",2),
                ("nova04",2),
                ("nova05",2),
                ("nova06",2),
                ("nova07",2),
                ("nova08",2),
                ("nova09",2),
                ("nova10",2),
                ("nova11",4),
                ("nova12",4),
                ("nova13",4),
                ("nova14",4),
                ],(x,y-t.h-28),(x,y-t.h-28),-1,delay=10),]
        return effects
    
class Red_Jelly(Unit):
    name = "Red Jelly"
    colour = "red"
    side = "jelly"
    
    max_hp = 60
    attack = 35
    defense = 12

    moverange = 4
    attackrange = 1
    targetsize = 0.5
    mode = "circle"
    shape = "circle"
    jump = 20
    def action(self,targets,victims,cursor,floor):
        vx,vy = targets[0]
        x,y = floor.data[(vx,vy)][-1].pos
        effects = [LineEffect([
            ("horn1",4),
            ("horn2",4),
            ("horn3",4),
            ("horn4",10),
            ],(x,y-64),(x,y-16),22)]
        return effects

class Green_Jelly(Unit):
    name = "Green Jelly"
    colour = "green"
    side = "jelly"
    
    max_hp = 40
    attack = -40
    defense = 5

    moverange = 5
    attackrange = 2.5
    targetmin = 0.5
    targetsize = 1
    mode = "circle"
    shape = "circle"
    jump = 20
    def action(self,targets,victims,cursor,floor):
        vx,vy = targets[0]
        x,y = floor.data[(vx,vy)][-1].pos
        effects = []
        return effects

class Yellow_Jelly(Unit):
    name = "Yellow Jelly"
    colour = "yellow"
    side = "jelly"
    
    max_hp = 50
    attack = 25
    defense = 6

    moverange = 5
    attackrange = 4
    targetsize = 0.5
    mode = "circle"
    shape = "circle"
    jump = 32

class Grey_Ooze(Unit):
    name = "Grey Ooze"
    colour = "grey"
    side = "ooze"
    
    max_hp = 50
    attack = 15
    defense = 8

    moverange = 5
    attackrange = 1
    targetsize = 0.5
    mode = "circle"
    shape = "circle"
    jump = 20
